/* ipwatch 
 *
 * ipwatch -- Look at the traffic going by on the ethernet.
 *
 * (1) Need to be root to use this.
 * (2) Uses etherd, nit, mgt, src
 * (3) Defaults to le0 (sparc default)
 * 
 * Written by:
 *
 * 1.0 thd@ornl.gov
 *
 * 1.3 ivan@ccs.neu.edu
 *     Ivan R Judson
 *     Northeastern University
 *     10/13/93
 * --Summary of Modifications--
 * Modified to do better command line parsing, as well as offer a 
 *  summary report of statistics gathered.  You can also specify which 
 *  interface you'd like to watch.
 * Put it into a real distribution, with makefile, man page, etc.
 *
 *  OkOkOk -- So I still haven't gotten to the man page, sigh....
 *     -- Ivan 3/16/95
 *
 */

#include "ipwatch.h"
#undef VERB1
#undef VERB

main(int argc, char **argv)
{
  void quit();
  int fd, packet_ip();
  int last, readbit, readfds, how_many;

  delay=5;
  
  if((argc > 3) || (argc == 2)) 
    printw("usage: ipwatch [-i interface]\n");
  else
    device = (argc == 3)?argv[2]:"le0";

  initdevice();
  initscr();

  signal(SIGINT, quit);
  signal(SIGTERM, quit);
  signal(SIGQUIT, quit);
  signal(SIGHUP, quit);

  last = time(0);
  readbit = 1 << if_fd;
  info = get_services();
#ifdef verb
  printf("%d", info->num_services);
#endif
  clear();
  refresh();

#ifdef verb
  currinfo = info->head->next;
  while (currinfo != info->head) {
    printf("service: %s\tprotocal: %s\tport: %d\n", currinfo->serv.s_name,
	   currinfo->serv.s_proto, currinfo->serv.s_port);
    currinfo = currinfo->next;
  }
  printf("number of services: %d\n", info->num_services);
#endif

  readfds = readbit;

  if (select(32, &readfds, 0, 0, 0) < 0) {
    if (errno != EINTR) {
      printw("ipwatch: select");
      quit();
    }
  }
  
  while(1){
    if (readfds & readbit) 
      readdata();

    if ((time(0) - last ) > delay) {
      update();
      last = time(0);
    }
  }
}

update()
{
  int percent, row, i, j, ndots, now, k;
  row = 0;
  
  now = time(0);
  move(row, 12);

  printw(" ipwatch-1.3\t  interface %s\t ippkts %5d\n", 
	 device, Pkts);
  if (Pkts == 0) {
    refresh(); return;
  }

  row = 1;
  move(row,0);  clrtoeol();
  printw("load: %4.0f kbs %d pps", KBs/(1000.*delay), AllPkts/delay);
  move(row, 25);
  percent = 100.*KBs/(1000.*delay*1200);
  printw("%3d%%",percent);      move(row, 30);
  ndots = percent/2;            while (ndots--) addch('*');
  move(++row, 0);               clrtoeol();
  printw(" arps");              move(row, 16);
  printw("%5d",arps);           move(row, 25);
  printw("%3d",arps/delay);     move(row, 30);
  ndots = arps/delay;
  if (ndots > 50) ndots = 50;
  while (ndots--) addch('*');
  move(++row, 0);               clrtoeol();
  printw(" ipmcast");           move(row, 16);
  printw("%5d",ipmcasts);       move(row, 25);
  printw("%3d",ipmcasts/delay); move(row, 30);
  ndots = ipmcasts/delay;
  if (ndots > 50) ndots=50;
  while (ndots--) addch('*');
  row = 3;

  currinfo = info->head;
  for(k=0;k<info->num_services;k++) { 
    if( currinfo->count > 0 ) { 
      move(++row, 0);
      clrtoeol();                 if (i) move(row,1);
      printw(" %s", currinfo->serv.s_name); move(row,16); 
      printw("%5d", currinfo->count);       move(row,25);
      ndots = 100.*currinfo->count/Pkts;
      printw("%3d%%", ndots); move(row,30);
      while (ndots--) addch('*');
    } 
    currinfo = currinfo->next;
  };

  currinfo = info->head;
  do {
    currinfo->count=0; 
    currinfo = currinfo->next;
  }while (currinfo != info->head);    

  refresh();
  arps = AllPkts = Pkts= ipmcasts = KBs = 0;
}


readdata()
{
  register struct nit_bufhdr *hdrp;
  register struct etherstat *p;
  register struct etheraddrs *q;
  caddr_t bp, bufstop;
  char *packet;
  int cc;
  
  if ((cc = read(if_fd, buf, sizeof buf)) < 0) {
    perror("ipwatch: read");
    exit(1);
  }
  
  for (bp = buf, bufstop = buf + cc; bp < bufstop;
       bp += hdrp->nhb_totlen) {

    hdrp = (struct nit_bufhdr *)bp;
    packet = (char *)(hdrp + 1);

    bcopy(packet, (char *)&eheader, sizeof (eheader));
    bcopy(packet + sizeof (eheader),  (char *)&netheader, sizeof (netheader));

    wirelen = hdrp->nhb_msglen;

    edst = (char *)&eheader.ether_dhost;
    esrc = (char *)&eheader.ether_shost;
    
    if (wirelen < MINPACKETLEN) {
      continue;
    }

    AllPkts++;
    KBs = KBs + wirelen;
    
    if (TYPE == ETHERTYPE_ARP)
      ; /* packet_arp(); */
    else
      if (TYPE == ETHERTYPE_IP)
	packet_ip(IP);
    else
      ;      /* packet_pup(); */
  }  
}

initdevice()
{
  struct strioctl si;
  struct ifreq    ifr;
  struct timeval  timeout;
  u_long          flags;
  
  if ((if_fd = open("/dev/nit", O_RDONLY)) < 0) {
    perror("nit open");
    exit (1);
  }
  
  /*
   * Arrange to get discrete messages from the stream.
   */
  if (ioctl(if_fd, I_SRDOPT, (char *)RMSGD) < 0) {
    perror("ioctl (I_SRDOPT)");
    exit (1);
  }
  
  si.ic_timout = INFTIM;
  
  /*
   * Push and configure the buffering module.
   */
  if (ioctl(if_fd, I_PUSH, "nbuf") < 0) {
    perror("ioctl (I_PUSH \"nbuf\")");
    exit (1);
  }
  
  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  si.ic_cmd = NIOCSTIME;
  si.ic_len = sizeof timeout;
  si.ic_dp = (char *)&timeout;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSTIME)");
    exit (1);
  }
  
  si.ic_cmd = NIOCSCHUNK;
  si.ic_len = sizeof chunksize;
  si.ic_dp = (char *)&chunksize;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSCHUNK)");
    exit (1);
  }
  
  /*
   * Configure the nit device, binding it to the proper
   * underlying interface and setting the interface into
   * promiscuous mode if necessary.
   */
  (void) strncpy(ifr.ifr_name, device, sizeof ifr.ifr_name);
  ifr.ifr_name[sizeof ifr.ifr_name - 1] = '\0';
  si.ic_cmd = NIOCBIND;
  si.ic_len = sizeof ifr;
  si.ic_dp = (char *)&ifr;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCBIND)");
    exit(1);
  }
  
  flags = NI_PROMISC;
  si.ic_cmd = NIOCSFLAGS;
  si.ic_len = sizeof flags;
  si.ic_dp = (char *)&flags;
  if (ioctl(if_fd, I_STR, (char *)&si) < 0) {
    perror("ioctl (I_STR: NIOCSFLAGS)");
    exit (1);
  }
  
  /*
   * Flush the read queue, to get rid of anything that
   * accumulated before the device reached its final configuration.
   */
  if (ioctl(if_fd, I_FLUSH, (char *)FLUSHR) < 0) {
    perror("ioctl (I_FLUSH)");
    exit(1);
  }
}

packet_ip(ip)
     register struct ip *ip;
{
  struct tcphdr *tp;
  struct udphdr *up;
  int i, k, fragmented, counter=0;

  Pkts++;
  tp = (struct tcphdr *)(ip + 1);
  up = (struct udphdr *)(ip + 1);
  fragmented = ip->ip_off & 0x1fff;  /*non zero offset */
  if (edst[0] == 1 && edst[1] == 0 && edst[2] == 0x5e ) ipmcasts++;
  
  switch (ip->ip_p) {
  case IPPROTO_TCP:
    currinfo = info->head;
    for(k=0;k<info->num_services;k++) { 
      if (tp->th_sport == currinfo->serv.s_port ||
	  tp->th_dport == currinfo->serv.s_port ) {
#ifdef VERB1
	printf("udp Port: %d count: %d\n", currinfo->serv.s_port, 
	       currinfo->count);
#endif
	currinfo->count = currinfo->count + 1;
	break;
      }
      currinfo = currinfo->next; counter++;
    }
    break;
  case IPPROTO_UDP:
    if (fragmented) up->uh_sport = 2049; /*nfs*/
    currinfo = info->head;
    for(k=0;k<info->num_services;k++) { 
      if (up->uh_sport == currinfo->serv.s_port ||
	  up->uh_dport == currinfo->serv.s_port ){
#ifdef VERB1
	printf("udp Port: %d count: %d\n", currinfo->serv.s_port, 
	       currinfo->count);
#endif
	currinfo->count = currinfo->count + 1;
	break;
      }
      currinfo = currinfo->next; counter++;
    }
    break;
  default:
    break;
  }
  return(1);
}

void quit()
{
  move(23,0);
  refresh();
  endwin();
  sleep(1);
  exit(0);
}
